"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const fs_1 = require("fs");
const promises_1 = require("fs/promises");
const stream_1 = require("stream");
const assertState_1 = __importDefault(require("./assertState"));
const rangeIntersection_1 = __importDefault(require("./rangeIntersection"));
exports.default = async ({ upstream, persistencePath, onPersisted, onChunkWriteError, }) => {
    const downstreams = [];
    // Offset of the next byte yet to be received.
    // This means we don't have byte at `next`, or any byte after it.
    let next = 0;
    let ended = false;
    // Use file handle instead of writable stream to ensure writes are flushed each time.
    const persistence = await promises_1.open(persistencePath, "w");
    (async () => {
        // Use async for loop to ensure chunks are processed one-by-one in order.
        for await (const chunk of upstream) {
            if (!Buffer.isBuffer(chunk)) {
                throw new TypeError(`Received non-Buffer data`);
            }
            const chunkStart = next;
            const chunkEnd = chunkStart + chunk.byteLength - 1;
            while (true) {
                try {
                    // Provide offset in case previous errors changed fh position.
                    await persistence.write(chunk, 0, chunk.byteLength, chunkStart);
                    break;
                }
                catch (e) {
                    if (!(await onChunkWriteError?.(e))) {
                        throw e;
                    }
                }
            }
            // `next` must be updated AFTER writing has been flushed to avoid race condition where a downstream is initialised based on a larger value of `next` but bytes haven't actually been written yet.
            next += chunk.byteLength;
            for (const d of downstreams) {
                if (d.next <= d.end) {
                    // Downstream still has pending bytes.
                    const range = rangeIntersection_1.default([chunkStart, chunkEnd], [d.next, d.end]);
                    if (range) {
                        d.initialised.then(() => d.stream.write(chunk.slice(range[0] - chunkStart, range[1] - chunkStart + 1)));
                        d.next = range[1] + 1;
                    }
                }
                else {
                    d.initialised.then(() => d.stream.end());
                }
            }
        }
        ended = true;
        onPersisted?.({ size: next });
        for (const d of downstreams) {
            d.initialised.then(() => d.stream.end());
        }
    })().finally(() => persistence.close());
    return {
        downstream(start = 0, end = Infinity) {
            if (ended) {
                return fs_1.createReadStream(persistencePath, { start, end });
            }
            // Cache `next` value.
            // We don't need to cache `writing` as code is synchronous up to usage of it.
            const persistedLen = next;
            assertState_1.default(start <= end);
            const stream = new stream_1.PassThrough();
            const downstreamNext = Math.max(start, persistedLen);
            downstreams.push({
                stream,
                end,
                next: downstreamNext,
                // We need to add this downstream immediately to `downstreams` as `next` could change if we await an asynchronous read from the file system.
                initialised: (async () => {
                    if (start >= persistedLen) {
                        return;
                    }
                    const fh = await promises_1.open(persistencePath, "r");
                    try {
                        const length = Math.min(end + 1, persistedLen) - start;
                        assertState_1.default(length > 0);
                        const buffer = Buffer.alloc(length);
                        const { bytesRead } = await fh.read({
                            buffer,
                            position: start,
                            length,
                        });
                        assertState_1.default(bytesRead === length, `Expected to read ${length} bytes from persisted file, but got ${bytesRead}`);
                        stream.write(buffer);
                    }
                    finally {
                        await fh.close();
                    }
                })().catch((err) => stream.destroy(new Error(`Encountered ${err.name} while initialising downstream: ${err.message}`))),
            });
            return stream;
        },
    };
};
